import yaml


class CyberpunkConfig:
    def __init__(
        self,
        audio_store: str = "local",
        storage_base_dir: str = "testdata/",
        s3_storage_bucket: str = "mybucket",
        s3_storage_base_dir: str = "audio/",
    ):
        # local | s3 | audius
        self.audio_store = audio_store

        self.local_storage = LocalStorageConfig(storage_base_dir)
        self.s3_storage = S3StorageConfig(
            s3_storage_bucket,
            s3_storage_base_dir,
        )

    @classmethod
    def from_yaml(cls, path="cyberpunk.yaml"):
        with open(path) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)

        # TODO: validation lol
        audio_source = data["audio_store"]
        storage_base_dir = data["local"]["storage_base_dir"]
        s3_storage_bucket = data["s3"]["s3_storage_bucket"]
        s3_storage_base_dir = data["s3"]["s3_storage_base_dir"]

        return cls(
            audio_source,
            storage_base_dir,
            s3_storage_bucket,
            s3_storage_base_dir,
        )


class LocalStorageConfig:
    def __init__(self, storage_base_dir: str = "testdata/"):
        self.storage_base_dir = storage_base_dir


class S3StorageConfig:
    def __init__(
        self,
        s3_storage_bucket: str = "mybucket",
        s3_storage_base_dir: str = "audio/",
    ):
        self.s3_storage_bucket = s3_storage_bucket
        self.s3_storage_base_dir = s3_storage_base_dir


# Cyberpunk Config Singleton
cyberpunk_config = CyberpunkConfig.from_yaml("cyberpunk.yaml")



from typing import Generator

from flask import Flask, Response, jsonify, request, stream_with_context

from cyberpunk.processing import parse_query, process_args


def create_app():
    app = Flask(__name__)

    def stream_audio_file(filename: str, chunk_size: int = 2048) -> Generator:
        with open(f"testdata/{filename}", "rb") as faudio:
            data = faudio.read(chunk_size)
            while data:
                yield data
                data = faudio.read(chunk_size)

    @app.route("/")
    def hello():
        return "Hello World"

    @app.route("/healthcheck")
    def healthcheck():
        return 200

    @app.route("/unsafe/<filename>", methods=["GET"])
    def unsafe_processing(filename: str):
        args = request.args
        processed_file, file_type = process_args(filename, args)

        return Response(
            stream_with_context(stream_audio_file(processed_file)),
            mimetype=file_type,
        )

    @app.route("/params/<filename>")
    def params_route(filename: str):
        return jsonify(parse_query(filename, request.args))

    return app



from typing import Any, Dict

from pydub import AudioSegment


class Slice(object):
    def __call__(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        return self.process(segment, inputs)

    def parse_input_from_str(self, arg: str) -> Dict:
        # TODO: Slices can have no start (implied 0) or no end (implied length of segment)

        start_str, end_str = tuple(arg.split(":"))

        start = int(start_str)
        end = int(end_str)

        return {
            "start": start,
            "end": end,
        }

    def process(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:

        start = inputs["start"]
        end = inputs["end"]

        sliced_segment = segment[start:end]
        return sliced_segment



from typing import Any, Dict

from pydub import AudioSegment


class Reverse(object):
    def __call__(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        return self.process(segment, inputs)

    def parse_input_from_str(self, arg: str) -> Dict:
        return {}

    def process(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        reversed_segment = segment.reverse()

        return reversed_segment



from typing import Any, Dict

from pydub import AudioSegment

from cyberpunk.storage import audio_storage


class Concat(object):
    def __call__(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        return self.process(segment, inputs)

    def parse_input_from_str(self, arg: str) -> Dict:
        other_filename = arg
        other_segment = audio_storage.get_segment(other_filename)

        return {
            "other": other_segment,
        }

    def process(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        other = inputs["other"]
        concated_segment = segment + other

        return concated_segment



from typing import Dict, Protocol, Type

from pydub import AudioSegment

from cyberpunk.config import cyberpunk_config
from cyberpunk.storage.audius import AudiusStorage
from cyberpunk.storage.local import LocalStorage
from cyberpunk.storage.s3 import S3Storage


class AudioStorage(Protocol):
    def __init__(self):
        """Declare variables like base dir"""

    def get_segment(self, key: str) -> AudioSegment:
        """get an audio segment from storage"""

    def save_segment(self, base_filename: str, segment: AudioSegment) -> str:
        """save an audio segment to storage and return the link to it"""


def configure_storage() -> AudioStorage:

    storage_table: Dict[str, Type[AudioStorage]] = {
        "local": LocalStorage,
        "s3": S3Storage,
        "audius": AudiusStorage,
    }
    store = cyberpunk_config.audio_store

    return storage_table[store]()


# Audio Storage Singleton
audio_storage: AudioStorage = configure_storage()



from typing import Any, Dict

from pydub import AudioSegment


class Repeat(object):
    def __call__(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        return self.process(segment, inputs)

    def parse_input_from_str(self, arg: str) -> Dict[str, Any]:
        multiplier = int(arg)

        return {
            "multiplier": multiplier,
        }

    def process(
        self,
        segment: AudioSegment,
        inputs: Dict[str, Any],
    ) -> AudioSegment:
        multiplier = inputs["multiplier"]
        repeated_segment = segment * multiplier

        return repeated_segment



import boto3
from pydub import AudioSegment


class S3Storage(object):
    def __init__(self):
        self.s3 = boto3.resource("s3")

    def get_segment(self, key: str) -> AudioSegment:
        pass

    def save_segment(
        self,
        base_filename: str,
        segment: AudioSegment,
    ) -> str:
        # TODO: export with Filename unique to the stages run (for caching)
        # TODO: All for exporting different file type (e.g. mp3, wav, etc.)
        processed_filename = f"processed_{base_filename}"
        segment.export(f"tmp/{processed_filename}", format="mp3")

        return processed_filename



import random

import requests
from pydub import AudioSegment


class AudiusStorage(object):
    def __init__(self):
        self.host = random.choice(
            (requests.get("https://api.audius.co")).json()["data"],
        )

    def get_segment(self, key: str) -> AudioSegment:
        req = requests.get(
            f"{self.host}/v1/tracks/{key}/stream",
            params={"app_name": "cyberpunk"},
        )

        with open(f"testdata/{key}.mp3", "wb") as f:
            f.write(req.content)

        segment = AudioSegment.from_file(f"testdata/{key}.mp3")

        return segment

    def save_segment(
        self,
        base_filename: str,
        segment: AudioSegment,
    ) -> str:
        # TODO: export with Filename unique to the stages run (for caching)
        # TODO: All for exporting different file type (e.g. mp3, wav, etc.)
        processed_filename = f"processed_{base_filename}"
        segment.export(f"tmp/{processed_filename}", format="mp3")

        return processed_filename



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cyberpunk', 'cyberpunk.storage', 'cyberpunk.transformations']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.2,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'boto3>=1.20.54,<2.0.0',
 'click>=8.0.3,<9.0.0',
 'gunicorn>=20.1.0,<21.0.0',
 'pydub>=0.25.1,<0.26.0',
 'python-dotenv>=0.19.2,<0.20.0',
 'requests>=2.27.1,<3.0.0']

entry_points = \
{'console_scripts': ['cyberpunk = main:main']}

setup_kwargs = {
    'name': 'cyberpunk',
    'version': '0.2.0',
    'description': 'Audio Processing Server',
    'long_description': '\n# Cyberpunk\n\nAudio Processing Server\n\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/jonaylor89/cyberpunk/Docker)\n\n[![Run on Google Cloud](https://deploy.cloud.run/button.svg)](https://deploy.cloud.run?git_repo=https://github.com/jonaylor89/cyberpunk)\n\n# Install\n\nWith Poetry\n```sh\ncurl -sSL https://install.python-poetry.org | python3 -\npoetry run main.py\n```\n\nWith Docker\n```\ndocker pull ghcr.io/jonaylor89/cyberpunk:sha256-36a2ec3c572d69b41a096cee07f60f2e07d669846ceb0dd8b9f54d741ecc678c.sig\ndocker run -e PORT=8080 jonaylor/cyberpunk\n```\n\n\n# Features\n\n- [x] Audio Streaming\n\n- [ ] Change encodings (e.g. mp3 -> wav)\n- [x] Audio slicing\n- [ ] Change Volume\n- [ ] Concat Audio\n- [x] Repeat Audio\n- [x] Reverse Audio\n- [ ] Crossfade\n- [ ] Fade in/out\n- [ ] Audio Quality\n- [ ] Audio Tagging\n- [ ] Audio Thumbnails\n- [ ] Mastering Music\n\n- [ ] Sound/Vocal Isolation\n\n- [ ] [Cool ML Stuff](https://github.com/spotify/pedalboard)\n\n- [ ] [File Caching](https://gist.github.com/ruanbekker/75d98a0d5cab5d6a562c70b4be5ba86d)\n\n# Storage Options\n\n- [x] Local\n- [ ] Cloud (e.g. S3)\n- [ ] Blockchain (Audius)\n\n\n# Environment\n\nCYBERPUNK_SECRET: mysecret # secret key for URL signature\n\nAWS_ACCESS_KEY_ID: ...\n\nAWS_SECRET_ACCESS_KEY: ...\n\nAWS_REGION: us-east-1',
    'author': 'Johannes Naylor',
    'author_email': 'jonaylor89@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonaylor89/cyberpunk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, Tuple

from pydub import AudioSegment

from cyberpunk.storage import audio_storage
from cyberpunk.transformations import (
    Concat,
    Repeat,
    Reverse,
    Slice,
    Transformation,
)


def process_args(base_filename: str, args: Dict) -> Tuple[str, str]:

    lookup_table: Dict[str, Transformation] = {
        "reverse": Reverse(),
        "repeat": Repeat(),
        "slice": Slice(),
        "concat": Concat(),
    }

    # Create Audio Segment
    audio_segment: AudioSegment = audio_storage.get_segment(base_filename)

    # Pass Audio Segment through Each Stage
    for (k, v) in args.items():
        if k in lookup_table.keys():
            transformation: Transformation = lookup_table[k]
            assert transformation is not None

            inputs: Dict[str, Any] = transformation.parse_input_from_str(v)
            audio_segment = transformation.process(audio_segment, inputs)

    processed_filename = audio_storage.save_segment(
        base_filename,
        audio_segment,
    )

    # Return Filename and Audio Type
    return processed_filename, "audio/mp3"


# TODO: This function should return the users input as json
# It'll help them debug
def parse_query(filename: str, args: Dict) -> Dict:
    return {
        "reverse": False,
        "repeat": 0,
        "slice": {
            "start": 0,
            "end": 0,
        },
    }
